@@ -1,11 +1,12 @@
 ;;; bufferlo.el --- Manage frame/tab-local buffer lists -*- lexical-binding: t -*-
+
 ;; Copyright (C) 2021-2023 Free Software Foundation, Inc.
 
 ;; Author: Florian Rommel <mail@florommel.de>
 ;; Maintainer: Florian Rommel <mail@florommel.de>
 ;; Url: https://github.com/florommel/bufferlo
 ;; Created: 2021-09-15
-;; Version: 0.2
+;; Version: 0.3
 ;; Package-Requires: ((emacs "27.1"))
 ;; Keywords: buffer frame tabs local
 
@@ -58,7 +59,7 @@
 (require 'desktop)
 
 (defgroup bufferlo nil
-  "Manage frame/tab local buffers."
+  "Manage frame/tab-local buffer lists."
   :group 'convenience)
 
 (defcustom bufferlo-prefer-local-buffers t
@@ -69,46 +70,43 @@ This is also required to make `next-buffer' and `previous-buffer'
 work as expected.
 Changes to this variable must be made before enabling
 `bufferlo-mode' in order to take effect."
-  :group 'bufferlo
   :type 'boolean)
 
 (defcustom bufferlo-include-buried-buffers t
   "Include buried buffers in the local list (`bufferlo-buffer-list').
 Use `bufferlo-bury' to remove and bury a buffer if this is set to t."
-  :group 'bufferlo
   :type 'boolean)
 
 (defcustom bufferlo-include-buffer-filters nil
   "Buffers that should always get included in a new tab or frame.
 This is a list of regular expressions that match buffer names.
+This is applied on frame and tab creation.  Included buffers can be
+explicitly removed later.
 This overrides buffers excluded by `bufferlo-exclude-buffer-filters.'"
-  :group 'bufferlo
   :type '(repeat string))
 
 (defcustom bufferlo-exclude-buffer-filters '(".*")
   "Buffers that should always get excluded in a new tab or frame.
 This is a list of regular expressions that match buffer names.
+This is applied on frame and tab creation.  Excluded buffers can be
+added explicitly later.  Use `bufferlo-hidden-buffers' to permanently
+hide buffers from the local list.
 Buffers included by `bufferlo-include-buffer-filters' take precedence."
-  :group 'bufferlo
   :type '(repeat string))
 
 (defcustom bufferlo-hidden-buffers nil
-  "Buffers that should be hidden in the local buffer lists,
-even if they are displayed in the current frame or tab.
-This is a list of regular expressions that match buffer names."
-  :group 'bufferlo
+  "List of regexps matching names of buffers to hide in the local buffer lists.
+Matching buffers are hidden even if displayed in the current frame or tab."
   :type '(repeat string))
 
 (defcustom bufferlo-kill-buffers-exclude-filters
   '("\\` " "\\`\\*Messages\\*\\'" "\\`\\*scratch\\*\\'")
   "Buffers that should not be killed by `bufferlo-kill-buffers'.
 This is a list of regular expressions that match buffer names."
-  :group 'bufferlo
   :type '(repeat string))
 
 (defcustom bufferlo-ibuffer-bind-local-buffer-filter t
-  "If this is true bind the local buffer filter to \"/ l\" in ibuffer."
-  :group 'bufferlo
+  "If non-nil, the local buffer filter is bound to \"/ l\" in ibuffer."
   :type '(repeat string))
 
 (defvar bufferlo--desktop-advice-active nil)
@@ -120,7 +118,6 @@ This is a list of regular expressions that match buffer names."
 (define-minor-mode bufferlo-mode
   "Manage frame/tab-local buffers."
   :global t
-  :group 'bufferlo
   :require 'bufferlo
   :init-value nil
   :lighter nil
@@ -143,8 +140,10 @@ This is a list of regular expressions that match buffer names."
         ;; Duplicate/move tabs
         (advice-add #'tab-bar-select-tab :around #'bufferlo--activate-force)
         ;; Clone & undelete frame
-        (advice-add #'clone-frame :around #'bufferlo--activate-force)
-        (advice-add #'undelete-frame :around #'bufferlo--activate-force)
+        (when (>= emacs-major-version 28)
+          (advice-add #'clone-frame :around #'bufferlo--activate-force))
+        (when (>= emacs-major-version 29)
+          (advice-add #'undelete-frame :around #'bufferlo--activate-force))
         ;; Switch-tab workaround
         (advice-add #'tab-bar-select-tab :around #'bufferlo--clear-buffer-lists-activate)
         (advice-add #'tab-bar--tab :after #'bufferlo--clear-buffer-lists))
@@ -163,14 +162,16 @@ This is a list of regular expressions that match buffer names."
     ;; Duplicate/move tabs
     (advice-remove #'tab-bar-select-tab #'bufferlo--activate-force)
     ;; Clone & undelete frame
-    (advice-remove #'clone-frame #'bufferlo--activate-force)
-    (advice-remove #'undelete-frame #'bufferlo--activate-force)
+    (when (>= emacs-major-version 28)
+      (advice-remove #'clone-frame #'bufferlo--activate-force))
+    (when (>= emacs-major-version 29)
+      (advice-remove #'undelete-frame #'bufferlo--activate-force))
     ;; Switch-tab workaround
     (advice-remove #'tab-bar-select-tab #'bufferlo--clear-buffer-lists-activate)
     (advice-remove #'tab-bar--tab #'bufferlo--clear-buffer-lists)))
 
 (defun bufferlo-local-buffer-p (buffer &optional frame tabnum include-hidden)
-  "Return whether BUFFER is in the list of local buffers.
+  "Return non-nil if BUFFER is in the list of local buffers.
 A non-nil value of FRAME selects a specific frame instead of the current one.
 If TABNUM is nil, the current tab is used.  If it is non-nil, it specifies
 a tab index in the given frame.  If INCLUDE-HIDDEN is set, include hidden
@@ -178,7 +179,7 @@ buffers, see `bufferlo-hidden-buffers'."
   (memq buffer (bufferlo-buffer-list frame tabnum include-hidden)))
 
 (defun bufferlo-non-local-buffer-p (buffer &optional frame tabnum include-hidden)
-  "Return whether BUFFER is not in the list of local buffers.
+  "Return non-nil if BUFFER is not in the list of local buffers.
 A non-nil value of FRAME selects a specific frame instead of the current one.
 If TABNUM is nil, the current tab is used.  If it is non-nil, it specifies
 a tab index in the given frame.  If INCLUDE-HIDDEN is set, include hidden
@@ -201,7 +202,7 @@ when `tab-bar--tab' is called from `tab-bar-select-tab."
     (set-frame-parameter frame 'buried-buffer-list nil)))
 
 (defun bufferlo--clear-buffer-lists-activate (oldfn &rest args)
-  "This should be set up as a device around `tab-bar-select-tab'.
+  "This should be set up as a advice around `tab-bar-select-tab'.
 It actiavtes clearing the buffer lists for `tab-bar--tab'
 before calling OLDFN with ARGS.  See `bufferlo--clear-buffer-lists'."
   (let* ((bufferlo--clear-buffer-lists-active t)
@@ -241,7 +242,7 @@ Includes hidden buffers."
              regexp-list "\\|"))
 
 (defun bufferlo--include-exclude-buffers (frame)
-  "Include and exclude buffers into the buffer list of FRAME."
+  "Include and exclude buffers from the local buffer list of FRAME."
   (let* ((include (bufferlo--merge-regexp-list
                    (append '("a^") bufferlo-include-buffer-filters)))
          (exclude (bufferlo--merge-regexp-list
@@ -266,7 +267,8 @@ Includes hidden buffers."
     (set-frame-parameter frame 'buried-buffer-list nil)))
 
 (defun bufferlo--tab-include-exclude-buffers (ignore)
-  "Include and exclude buffers into buffer-list of the current tab's frame."
+  "Include and exclude buffers from the buffer list of the current tab's frame.
+Argument IGNORE is for compatibility with `tab-bar-tab-post-open-functions'."
   (ignore ignore)
   ;; Reset the local buffer list unless we clone the tab (tab-duplicate).
   (unless (or (eq tab-bar-new-tab-choice 'clone)
@@ -318,14 +320,19 @@ buffers, see `bufferlo-hidden-buffers'."
                   list))))
 
 (defun bufferlo--window-state-get (oldfn &optional window writable)
-  "Save the frame's buffer-list to the window state."
+  "Save the frame's buffer list to the window state.
+Used as advice around `window-state-get'.  OLDFN is the original
+function.  WINDOW and WRITABLE are passed to the function."
   (let ((ws (apply oldfn (list window writable))))
     (let* ((buffers (bufferlo--current-buffers (window-frame window)))
            (names (mapcar #'buffer-name buffers)))
       (if names (append ws (list (list 'bufferlo-buffer-list names))) ws))))
 
 (defun bufferlo--window-state-put (state &optional window ignore)
-  "Restore the frame's buffer-list from the window state."
+  "Restore the frame's buffer list from the window state.
+Used as advice after `window-state-put'.  STATE is the window state.
+WINDOW is the window in question.  IGNORE is not used and exists for
+compatibility with the adviced function."
   (ignore ignore)
   ;; We have to make sure that the window is live at this point.
   ;; `frameset-restore' may pass a window with a non-existing buffer
@@ -346,12 +353,16 @@ buffers, see `bufferlo-hidden-buffers'."
                            (list (window-buffer window))))))
 
 (defun bufferlo--activate (oldfn &rest args)
-  "Activate the advice for `bufferlo--window-state-{get,put}'."
+  "Activate the advice for `bufferlo--window-state-{get,put}'.
+OLDFN is the original function.  ARGS is for compatibility with
+the adviced functions."
   (let ((bufferlo--desktop-advice-active t))
     (apply oldfn args)))
 
 (defun bufferlo--activate-force (oldfn &rest args)
-  "Activate the advice for `bufferlo--window-state-{get,put}'."
+  "Activate the advice for `bufferlo--window-state-{get,put}'.
+OLDFN is the original function.  ARGS is for compatibility with
+the adviced functions."
   (let ((bufferlo--desktop-advice-active t)
         (bufferlo--desktop-advice-active-force t))
     (apply oldfn args)))
@@ -394,8 +405,7 @@ If FRAME is nil, use the current frame."
     nil))
 
 (defun bufferlo-remove-non-exclusive-buffers ()
-  "Remove all buffers from the frame/tab's buffer list that are not
-exclusively attached to this frame/tab."
+  "Remove all buffers from the local buffer list that are not exclusive to it."
   (interactive)
   (dolist (buffer (bufferlo--get-exclusive-buffers nil t))
     (bufferlo-remove buffer)))
@@ -439,9 +449,9 @@ If EXCLUDE-FRAME is a frame, exclude the local buffer list of this frame."
               (buffer-list)))
 
 (defun bufferlo--get-exclusive-buffers (&optional frame invert)
-  "Get all buffers that are exclusive for this frame and tab.
+  "Get all buffers that are exclusive to this frame and tab.
 If FRAME is nil, use the current frame.
-If INVERT is non-nil, return the non-exclusive buffer instead."
+If INVERT is non-nil, return the non-exclusive buffers instead."
   (let ((other-bufs (bufferlo--get-captured-buffers (or frame (selected-frame))))
         (this-bufs (bufferlo--current-buffers frame)))
     (seq-filter (if invert
@@ -477,21 +487,38 @@ Buffers matching `bufferlo-kill-buffers-exclude-filters' are never killed."
         (kill-buffer buffer)))))
 
 (defun bufferlo-delete-frame-kill-buffers (&optional frame)
-  "Delete a frame and kill all the local buffers according to
-`bufferlo-kill-buffers'.
+  "Delete a frame and kill the local buffers.
 If FRAME is nil, kill the current frame."
   (interactive)
   (bufferlo-kill-buffers frame)
   (delete-frame))
 
 (defun bufferlo-tab-close-kill-buffers (&optional killall)
-  "Close the current tab and kill all the local buffers according to
-`bufferlo-kill-buffers'.
+  "Close the current tab and kill the local buffers.
 The optional parameter KILLALL is passed to `bufferlo-kill-buffers'"
   (interactive "P")
   (bufferlo-kill-buffers killall)
   (tab-bar-close-tab))
 
+(defun bufferlo-isolate-project (&optional file-buffers-only)
+  "Isolate a project in the frame or tab.
+Remove all buffers that do not belong to the current project from
+the local buffer list.  When FILE-BUFFERS-ONLY is non-nil or the
+prefix argument is given, remove only buffers that visit a file.
+Buffers matching `bufferlo-include-buffer-filters' are not removed."
+  (interactive "P")
+  (let ((curr-project (project-current))
+        (include (bufferlo--merge-regexp-list
+                  (append '("a^") bufferlo-include-buffer-filters))))
+    (if curr-project
+        (dolist (buffer (bufferlo-buffer-list))
+          (when (and (not (string-match-p include (buffer-name buffer)))
+                     (not (equal curr-project
+                                 (with-current-buffer buffer (project-current))))
+                     (or (not file-buffers-only) (buffer-file-name buffer)))
+            (bufferlo-remove buffer)))
+      (message "Current buffer is not part of a project"))))
+
 (defun bufferlo-switch-to-buffer (buffer &optional norecord force-same-window)
   "Display the BUFFER in the selected window.
 Completion includes only local buffers.
@@ -508,11 +535,35 @@ If the prefix arument is given, include all buffers."
          (lambda (b) (member (if (stringp b) b (car b)) lbs)))))))
   (switch-to-buffer buffer norecord force-same-window))
 
+(defvar-local bufferlo--buffer-menu-this-frame nil)
+
+(defun bufferlo--local-buffer-list-this-frame ()
+  "Return the local buffer list of the buffer's frame."
+  (bufferlo-buffer-list bufferlo--buffer-menu-this-frame))
+
 (defun bufferlo-list-buffers ()
-  "Display a list of local buffers in the \"*Buffer List*\" buffer."
+  "Display a list of local buffers."
   (interactive)
   (display-buffer
-   (list-buffers-noselect nil #'bufferlo-buffer-list)))
+   (let* ((old-buffer (current-buffer))
+          (name (or
+                 (seq-find (lambda (b)
+                             (string-match-p
+                              "\\`\\*Local Buffer List\\*\\(<[0-9]*>\\)?\\'"
+                              (buffer-name b)))
+                           (bufferlo-buffer-list))
+                 (generate-new-buffer-name "*Local Buffer List*")))
+	  (buffer (get-buffer-create name)))
+     (with-current-buffer buffer
+       (Buffer-menu-mode)
+       (setq bufferlo--buffer-menu-this-frame (selected-frame))
+       (setq Buffer-menu-files-only nil)
+       (setq Buffer-menu-buffer-list #'bufferlo--local-buffer-list-this-frame)
+       (setq Buffer-menu-filter-predicate nil)
+       (list-buffers--refresh #'bufferlo--local-buffer-list-this-frame old-buffer)
+       (tabulated-list-print)
+       (revert-buffer))
+     buffer)))
 
 (with-eval-after-load 'ibuf-ext
   (define-ibuffer-filter bufferlo-local-buffers
